#include <msp/core/algorithm.h>
#include <msp/strings/format.h>
#include <msp/strings/utils.h>
#include "builtin.h"
#include "compatibility.h"
#include "compiler.h"
#include "debug.h"
#include "error.h"
#include "generate.h"
#include "glsl_error.h"
#include "optimize.h"
#include "output.h"
#include "resources.h"
#include "validate.h"

#undef interface

using namespace std;

namespace Msp {
namespace GL {
namespace SL {

Compiler::Compiler():
	features(Features::from_context()),
	module(0),
	specialized(false)
{ }

Compiler::Compiler(const Features &f):
	features(f),
	module(0),
	specialized(false)
{ }

Compiler::~Compiler()
{
	delete module;
}

void Compiler::clear()
{
	delete module;
	module = new Module();
	imported_names.clear();
	module->source_map.set_name(0, "<generated>");
}

void Compiler::set_source(const string &source, const string &src_name)
{
	clear();
	Parser parser;
	imported_names.push_back(src_name);
	append_module(parser.parse(source, src_name, 1), 0);
}

void Compiler::load_source(IO::Base &io, DataFile::Collection *res, const string &src_name)
{
	clear();
	Parser parser;
	imported_names.push_back(src_name);
	append_module(parser.parse(io, src_name, 1), res);
}

void Compiler::load_source(IO::Base &io, const string &src_name)
{
	load_source(io, 0, src_name);
}

void Compiler::specialize(const map<string, int> &sv)
{
	specialized = true;
	spec_values = sv;
}

void Compiler::compile(Mode mode)
{
	for(list<Stage>::iterator i=module->stages.begin(); i!=module->stages.end(); ++i)
		generate(*i, mode);

	bool valid = true;
	for(list<Stage>::iterator i=module->stages.begin(); i!=module->stages.end(); ++i)
		if(!validate(*i))
			valid = false;
	if(!valid)
		throw invalid_shader_source(get_diagnostics());

	unsigned n = 0;
	for(list<Stage>::iterator i=module->stages.begin(); i!=module->stages.end(); ++n)
	{
		OptimizeResult result = optimize(*i);
		if(result==REDO_PREVIOUS)
			i = module->stages.begin();
		else if(result!=REDO_STAGE)
			++i;
	}
	for(list<Stage>::iterator i=module->stages.begin(); i!=module->stages.end(); ++i)
		finalize(*i, mode);
}

string Compiler::get_combined_glsl() const
{
	string glsl;

	unsigned source_count = module->source_map.get_count();
	for(unsigned i=1; i<source_count; ++i)
		glsl += format("#pragma MSP source(%d, \"%s\")\n", i, module->source_map.get_name(i));
	for(list<Stage>::iterator i=module->stages.begin(); i!=module->stages.end(); ++i)
	{
		glsl += format("#pragma MSP stage(%s)\n", Stage::get_stage_name(i->type));
		glsl += Formatter().apply(*i);
		glsl += '\n';
	}

	return glsl;
}

vector<Stage::Type> Compiler::get_stages() const
{
	vector<Stage::Type> stage_types;
	stage_types.reserve(module->stages.size());
	for(list<Stage>::const_iterator i=module->stages.begin(); i!=module->stages.end(); ++i)
		stage_types.push_back(i->type);
	return stage_types;
}

string Compiler::get_stage_glsl(Stage::Type stage_type) const
{
	for(list<Stage>::iterator i=module->stages.begin(); i!=module->stages.end(); ++i)
		if(i->type==stage_type)
			return Formatter().apply(*i);
	throw key_error(Stage::get_stage_name(stage_type));
}

const map<string, unsigned> &Compiler::get_vertex_attributes() const
{
	for(list<Stage>::const_iterator i=module->stages.begin(); i!=module->stages.end(); ++i)
		if(i->type==Stage::VERTEX)
			return i->locations;
	throw invalid_operation("Compiler::get_vertex_attributes");
}

const map<string, unsigned> &Compiler::get_fragment_outputs() const
{
	for(list<Stage>::const_iterator i=module->stages.begin(); i!=module->stages.end(); ++i)
		if(i->type==Stage::FRAGMENT)
			return i->locations;
	throw invalid_operation("Compiler::get_fragment_outputs");
}

const SourceMap &Compiler::get_source_map() const
{
	return module->source_map;
}

string Compiler::get_stage_debug(Stage::Type stage_type) const
{
	for(list<Stage>::iterator i=module->stages.begin(); i!=module->stages.end(); ++i)
		if(i->type==stage_type)
			return DumpTree().apply(*i);
	throw key_error(Stage::get_stage_name(stage_type));
}

string Compiler::get_diagnostics() const
{
	string combined;
	for(list<Stage>::const_iterator i=module->stages.begin(); i!=module->stages.end(); ++i)
		for(vector<Diagnostic>::const_iterator j=i->diagnostics.begin(); j!=i->diagnostics.end(); ++j)
			if(j->source!=INTERNAL_SOURCE)
				append(combined, "\n", format("%s:%d: %s", module->source_map.get_name(j->source), j->line, j->message));
	return combined;
}

void Compiler::append_module(Module &mod, DataFile::Collection *res)
{
	module->source_map.merge_from(mod.source_map);

	vector<Import *> imports;
	for(NodeList<Statement>::const_iterator i=mod.shared.content.body.begin(); i!=mod.shared.content.body.end(); ++i)
		if(Import *imp = dynamic_cast<Import *>(i->get()))
			imports.push_back(imp);
	for(vector<Import *>::iterator i=imports.begin(); i!=imports.end(); ++i)
		import(res, (*i)->module);
	NodeRemover().apply(mod.shared, set<Node *>(imports.begin(), imports.end()));

	append_stage(mod.shared);
	for(list<Stage>::iterator i=mod.stages.begin(); i!=mod.stages.end(); ++i)
		append_stage(*i);
}

void Compiler::append_stage(Stage &stage)
{
	Stage *target = 0;
	if(stage.type==Stage::SHARED)
		target = &module->shared;
	else
	{
		list<Stage>::iterator i;
		for(i=module->stages.begin(); (i!=module->stages.end() && i->type<stage.type); ++i) ;
		if(i==module->stages.end() || i->type>stage.type)
		{
			list<Stage>::iterator j = module->stages.insert(i, stage.type);
			if(i!=module->stages.end())
				i->previous = &*j;
			i = j;
			if(i!=module->stages.begin())
				i->previous = &*--j;
		}

		target = &*i;
	}

	if(stage.required_features.glsl_version>target->required_features.glsl_version)
		target->required_features.glsl_version = stage.required_features.glsl_version;
	for(NodeList<Statement>::iterator i=stage.content.body.begin(); i!=stage.content.body.end(); ++i)
		target->content.body.push_back(*i);
}

void Compiler::import(DataFile::Collection *resources, const string &name)
{
	string fn = name+".glsl";
	if(find(imported_names, fn)!=imported_names.end())
		return;
	imported_names.push_back(fn);

	RefPtr<IO::Seekable> io = (resources ? resources->open_raw(fn) : Resources::get_builtins().open(fn));
	if(!io)
		throw runtime_error(format("module %s not found", name));
	Parser import_parser;
	append_module(import_parser.parse(*io, fn, module->source_map.get_count()), resources);
}

void Compiler::generate(Stage &stage, Mode mode)
{
	stage.required_features.gl_api = features.gl_api;
	if(module->shared.required_features.glsl_version>stage.required_features.glsl_version)
		stage.required_features.glsl_version = module->shared.required_features.glsl_version;

	inject_block(stage.content, module->shared.content);
	if(const Stage *builtins = get_builtins(stage.type))
		inject_block(stage.content, builtins->content);
	if(const Stage *builtins = get_builtins(Stage::SHARED))
		inject_block(stage.content, builtins->content);

	// Initial resolving pass
	resolve(stage);

	/* All variables local to a stage have been resolved.  Resolve non-local
	variables through interfaces. */
	InterfaceGenerator().apply(stage);
	resolve(stage, RESOLVE_BLOCKS|RESOLVE_TYPES|RESOLVE_VARIABLES);

	ConstantSpecializer().apply(stage, (mode==PROGRAM && specialized ? &spec_values : 0));
	if(mode==PROGRAM)
	{
		LegacyConverter().apply(stage, features);
		resolve(stage, RESOLVE_VARIABLES|RESOLVE_FUNCTIONS);
	}
}

template<typename T>
bool Compiler::resolve(Stage &stage, unsigned &flags, unsigned bit)
{
	if(!(flags&bit))
		return false;

	flags &= ~bit;
	return T().apply(stage);
}

void Compiler::resolve(Stage &stage, unsigned flags)
{
	while(flags)
	{
		if(resolve<BlockHierarchyResolver>(stage, flags, RESOLVE_BLOCKS))
			;
		else if(resolve<TypeResolver>(stage, flags, RESOLVE_TYPES))
			flags |= RESOLVE_BLOCKS|RESOLVE_VARIABLES|RESOLVE_EXPRESSIONS;
		else if(resolve<VariableResolver>(stage, flags, RESOLVE_VARIABLES))
			flags |= RESOLVE_EXPRESSIONS;
		else if(resolve<FunctionResolver>(stage, flags, RESOLVE_FUNCTIONS))
			flags |= RESOLVE_EXPRESSIONS;
		else if(resolve<ExpressionResolver>(stage, flags, RESOLVE_EXPRESSIONS))
			flags |= RESOLVE_VARIABLES|RESOLVE_FUNCTIONS;
	}
}

bool Compiler::validate(Stage &stage)
{
	TypeValidator().apply(stage);
	DeclarationValidator().apply(stage);
	ReferenceValidator().apply(stage);
	ExpressionValidator().apply(stage);

	stable_sort(stage.diagnostics, &diagnostic_line_order);

	for(vector<Diagnostic>::const_iterator i=stage.diagnostics.begin(); i!=stage.diagnostics.end(); ++i)
		if(i->severity==Diagnostic::ERR)
			return false;

	return true;
}

bool Compiler::diagnostic_line_order(const Diagnostic &diag1, const Diagnostic &diag2)
{
	if(diag1.provoking_source!=diag2.provoking_source)
	{
		// Sort builtins first and imported modules according to import order.
		if(diag1.provoking_source<=BUILTIN_SOURCE)
			return diag1.provoking_source<diag2.provoking_source;
		else if(diag2.provoking_source<=BUILTIN_SOURCE)
			return false;
		else
			return diag1.provoking_source>diag2.provoking_source;
	}
	return diag1.provoking_line<diag2.provoking_line;
}

Compiler::OptimizeResult Compiler::optimize(Stage &stage)
{
	ConstantConditionEliminator().apply(stage);

	bool any_inlined = false;
	if(FunctionInliner().apply(stage))
	{
		resolve(stage, RESOLVE_TYPES|RESOLVE_VARIABLES|RESOLVE_FUNCTIONS|RESOLVE_EXPRESSIONS);
		any_inlined = true;
	}
	if(ExpressionInliner().apply(stage))
	{
		resolve(stage, RESOLVE_VARIABLES|RESOLVE_FUNCTIONS|RESOLVE_EXPRESSIONS);
		any_inlined = true;
	}

	/* Removing variables or functions may cause things from the previous stage
	to become unused. */
	bool any_removed = UnusedVariableRemover().apply(stage);
	any_removed |= UnusedFunctionRemover().apply(stage);
	any_removed |= UnusedTypeRemover().apply(stage);

	return any_removed ? REDO_PREVIOUS : any_inlined ? REDO_STAGE : NEXT_STAGE;
}

void Compiler::finalize(Stage &stage, Mode mode)
{
	if(get_gl_api()==OPENGL_ES2 && mode==PROGRAM)
		DefaultPrecisionGenerator().apply(stage);
	else if(mode==MODULE)
		PrecisionRemover().apply(stage);
}

void Compiler::inject_block(Block &target, const Block &source)
{
	NodeList<Statement>::iterator insert_point = target.body.begin();
	for(NodeList<Statement>::const_iterator i=source.body.begin(); i!=source.body.end(); ++i)
		target.body.insert(insert_point, (*i)->clone());
}

} // namespace SL
} // namespace GL
} // namespace Msp
